/* Automatically generated by tzextract */

#include "TimeZoneNameMap.h"

static const TimeZoneNameMapEntry TimeZoneNameMap[] = {
	{ "Dateline Standard Time", "Dateline Standard Time",
	  "(UTC-12:00) International Date Line West", "Dateline Daylight Time", "Etc/GMT+12" },
	{ "UTC-11", "UTC-11", "(UTC-11:00) Coordinated Universal Time-11", "UTC-11", "Etc/GMT+11" },
	{ "Aleutian Standard Time", "Aleutian Standard Time", "(UTC-10:00) Aleutian Islands",
	  "Aleutian Daylight Time", "America/Adak" },
	{ "Hawaiian Standard Time", "Hawaiian Standard Time", "(UTC-10:00) Hawaii",
	  "Hawaiian Daylight Time", "Pacific/Honolulu" },
	{ "Marquesas Standard Time", "Marquesas Standard Time", "(UTC-09:30) Marquesas Islands",
	  "Marquesas Daylight Time", "Pacific/Marquesas" },
	{ "Alaskan Standard Time", "Alaskan Standard Time", "(UTC-09:00) Alaska",
	  "Alaskan Daylight Time", "America/Anchorage" },
	{ "UTC-09", "UTC-09", "(UTC-09:00) Coordinated Universal Time-09", "UTC-09", "Etc/GMT+9" },
	{ "Pacific Standard Time (Mexico)", "Pacific Standard Time (Mexico)",
	  "(UTC-08:00) Baja California", "Pacific Daylight Time (Mexico)", "America/Tijuana" },
	{ "UTC-08", "UTC-08", "(UTC-08:00) Coordinated Universal Time-08", "UTC-08", "Etc/GMT+8" },
	{ "Pacific Standard Time", "Pacific Standard Time", "(UTC-08:00) Pacific Time (US & Canada)",
	  "Pacific Daylight Time", "America/Los_Angeles" },
	{ "US Mountain Standard Time", "US Mountain Standard Time", "(UTC-07:00) Arizona",
	  "US Mountain Daylight Time", "America/Phoenix" },
	{ "Mountain Standard Time (Mexico)", "Mountain Standard Time (Mexico)",
	  "(UTC-07:00) La Paz, Mazatlan", "Mountain Daylight Time (Mexico)", "America/Chihuahua" },
	{ "Mountain Standard Time", "Mountain Standard Time", "(UTC-07:00) Mountain Time (US & Canada)",
	  "Mountain Daylight Time", "America/Denver" },
	{ "Yukon Standard Time", "Yukon Standard Time", "(UTC-07:00) Yukon", "Yukon Daylight Time",
	  "America/Whitehorse" },
	{ "Central America Standard Time", "Central America Standard Time",
	  "(UTC-06:00) Central America", "Central America Daylight Time", "America/Guatemala" },
	{ "Central Standard Time", "Central Standard Time", "(UTC-06:00) Central Time (US & Canada)",
	  "Central Daylight Time", "America/Chicago" },
	{ "Easter Island Standard Time", "Easter Island Standard Time", "(UTC-06:00) Easter Island",
	  "Easter Island Daylight Time", "Pacific/Easter" },
	{ "Central Standard Time (Mexico)", "Central Standard Time (Mexico)",
	  "(UTC-06:00) Guadalajara, Mexico City, Monterrey", "Central Daylight Time (Mexico)",
	  "America/Mexico_City" },
	{ "Canada Central Standard Time", "Canada Central Standard Time", "(UTC-06:00) Saskatchewan",
	  "Canada Central Daylight Time", "America/Regina" },
	{ "SA Pacific Standard Time", "SA Pacific Standard Time",
	  "(UTC-05:00) Bogota, Lima, Quito, Rio Branco", "SA Pacific Daylight Time", "America/Bogota" },
	{ "Eastern Standard Time (Mexico)", "Eastern Standard Time (Mexico)", "(UTC-05:00) Chetumal",
	  "Eastern Daylight Time (Mexico)", "America/Cancun" },
	{ "Eastern Standard Time", "Eastern Standard Time", "(UTC-05:00) Eastern Time (US & Canada)",
	  "Eastern Daylight Time", "America/New_York" },
	{ "Haiti Standard Time", "Haiti Standard Time", "(UTC-05:00) Haiti", "Haiti Daylight Time",
	  "America/Port-au-Prince" },
	{ "Cuba Standard Time", "Cuba Standard Time", "(UTC-05:00) Havana", "Cuba Daylight Time",
	  "America/Havana" },
	{ "US Eastern Standard Time", "US Eastern Standard Time", "(UTC-05:00) Indiana (East)",
	  "US Eastern Daylight Time", "America/Indianapolis" },
	{ "Turks And Caicos Standard Time", "Turks and Caicos Standard Time",
	  "(UTC-05:00) Turks and Caicos", "Turks and Caicos Daylight Time", "America/Grand_Turk" },
	{ "Paraguay Standard Time", "Paraguay Standard Time", "(UTC-04:00) Asuncion",
	  "Paraguay Daylight Time", "America/Asuncion" },
	{ "Atlantic Standard Time", "Atlantic Standard Time", "(UTC-04:00) Atlantic Time (Canada)",
	  "Atlantic Daylight Time", "America/Halifax" },
	{ "Venezuela Standard Time", "Venezuela Standard Time", "(UTC-04:00) Caracas",
	  "Venezuela Daylight Time", "America/Caracas" },
	{ "Central Brazilian Standard Time", "Central Brazilian Standard Time", "(UTC-04:00) Cuiaba",
	  "Central Brazilian Daylight Time", "America/Cuiaba" },
	{ "SA Western Standard Time", "SA Western Standard Time",
	  "(UTC-04:00) Georgetown, La Paz, Manaus, San Juan", "SA Western Daylight Time",
	  "America/La_Paz" },
	{ "Pacific SA Standard Time", "Pacific SA Standard Time", "(UTC-04:00) Santiago",
	  "Pacific SA Daylight Time", "America/Santiago" },
	{ "Newfoundland Standard Time", "Newfoundland Standard Time", "(UTC-03:30) Newfoundland",
	  "Newfoundland Daylight Time", "America/St_Johns" },
	{ "Tocantins Standard Time", "Tocantins Standard Time", "(UTC-03:00) Araguaina",
	  "Tocantins Daylight Time", "America/Araguaina" },
	{ "E. South America Standard Time", "E. South America Standard Time", "(UTC-03:00) Brasilia",
	  "E. South America Daylight Time", "America/Sao_Paulo" },
	{ "SA Eastern Standard Time", "SA Eastern Standard Time", "(UTC-03:00) Cayenne, Fortaleza",
	  "SA Eastern Daylight Time", "America/Cayenne" },
	{ "Argentina Standard Time", "Argentina Standard Time", "(UTC-03:00) City of Buenos Aires",
	  "Argentina Daylight Time", "America/Buenos_Aires" },
	{ "Montevideo Standard Time", "Montevideo Standard Time", "(UTC-03:00) Montevideo",
	  "Montevideo Daylight Time", "America/Montevideo" },
	{ "Magallanes Standard Time", "Magallanes Standard Time", "(UTC-03:00) Punta Arenas",
	  "Magallanes Daylight Time", "America/Punta_Arenas" },
	{ "Saint Pierre Standard Time", "Saint Pierre Standard Time",
	  "(UTC-03:00) Saint Pierre and Miquelon", "Saint Pierre Daylight Time", "America/Miquelon" },
	{ "Bahia Standard Time", "Bahia Standard Time", "(UTC-03:00) Salvador", "Bahia Daylight Time",
	  "America/Bahia" },
	{ "UTC-02", "UTC-02", "(UTC-02:00) Coordinated Universal Time-02", "UTC-02", "Etc/GMT+2" },
	{ "Greenland Standard Time", "Greenland Standard Time", "(UTC-02:00) Greenland",
	  "Greenland Daylight Time", "America/Godthab" },
	{ "Mid-Atlantic Standard Time", "Mid-Atlantic Standard Time", "(UTC-02:00) Mid-Atlantic - Old",
	  "Mid-Atlantic Daylight Time", "" },
	{ "Azores Standard Time", "Azores Standard Time", "(UTC-01:00) Azores", "Azores Daylight Time",
	  "Atlantic/Azores" },
	{ "Cape Verde Standard Time", "Cabo Verde Standard Time", "(UTC-01:00) Cabo Verde Is.",
	  "Cabo Verde Daylight Time", "Atlantic/Cape_Verde" },
	{ "UTC", "Coordinated Universal Time", "(UTC) Coordinated Universal Time",
	  "Coordinated Universal Time", "Etc/UTC" },
	{ "GMT Standard Time", "GMT Standard Time", "(UTC+00:00) Dublin, Edinburgh, Lisbon, London",
	  "GMT Daylight Time", "Europe/London" },
	{ "Greenwich Standard Time", "Greenwich Standard Time", "(UTC+00:00) Monrovia, Reykjavik",
	  "Greenwich Daylight Time", "Atlantic/Reykjavik" },
	{ "Sao Tome Standard Time", "Sao Tome Standard Time", "(UTC+00:00) Sao Tome",
	  "Sao Tome Daylight Time", "Africa/Sao_Tome" },
	{ "Morocco Standard Time", "Morocco Standard Time", "(UTC+01:00) Casablanca",
	  "Morocco Daylight Time", "Africa/Casablanca" },
	{ "W. Europe Standard Time", "W. Europe Standard Time",
	  "(UTC+01:00) Amsterdam, Berlin, Bern, Rome, Stockholm, Vienna", "W. Europe Daylight Time",
	  "Europe/Berlin" },
	{ "Central Europe Standard Time", "Central Europe Standard Time",
	  "(UTC+01:00) Belgrade, Bratislava, Budapest, Ljubljana, Prague",
	  "Central Europe Daylight Time", "Europe/Budapest" },
	{ "Romance Standard Time", "Romance Standard Time",
	  "(UTC+01:00) Brussels, Copenhagen, Madrid, Paris", "Romance Daylight Time", "Europe/Paris" },
	{ "Central European Standard Time", "Central European Standard Time",
	  "(UTC+01:00) Sarajevo, Skopje, Warsaw, Zagreb", "Central European Daylight Time",
	  "Europe/Warsaw" },
	{ "W. Central Africa Standard Time", "W. Central Africa Standard Time",
	  "(UTC+01:00) West Central Africa", "W. Central Africa Daylight Time", "Africa/Lagos" },
	{ "GTB Standard Time", "GTB Standard Time", "(UTC+02:00) Athens, Bucharest",
	  "GTB Daylight Time", "Europe/Bucharest" },
	{ "Middle East Standard Time", "Middle East Standard Time", "(UTC+02:00) Beirut",
	  "Middle East Daylight Time", "Asia/Beirut" },
	{ "Egypt Standard Time", "Egypt Standard Time", "(UTC+02:00) Cairo", "Egypt Daylight Time",
	  "Africa/Cairo" },
	{ "E. Europe Standard Time", "E. Europe Standard Time", "(UTC+02:00) Chisinau",
	  "E. Europe Daylight Time", "Europe/Chisinau" },
	{ "West Bank Standard Time", "West Bank Gaza Standard Time", "(UTC+02:00) Gaza, Hebron",
	  "West Bank Gaza Daylight Time", "Asia/Hebron" },
	{ "South Africa Standard Time", "South Africa Standard Time", "(UTC+02:00) Harare, Pretoria",
	  "South Africa Daylight Time", "Africa/Johannesburg" },
	{ "FLE Standard Time", "FLE Standard Time",
	  "(UTC+02:00) Helsinki, Kyiv, Riga, Sofia, Tallinn, Vilnius", "FLE Daylight Time",
	  "Europe/Kiev" },
	{ "Israel Standard Time", "Jerusalem Standard Time", "(UTC+02:00) Jerusalem",
	  "Jerusalem Daylight Time", "Asia/Jerusalem" },
	{ "South Sudan Standard Time", "South Sudan Standard Time", "(UTC+02:00) Juba",
	  "South Sudan Daylight Time", "" },
	{ "Kaliningrad Standard Time", "Russia TZ 1 Standard Time", "(UTC+02:00) Kaliningrad",
	  "Russia TZ 1 Daylight Time", "Europe/Kaliningrad" },
	{ "Sudan Standard Time", "Sudan Standard Time", "(UTC+02:00) Khartoum", "Sudan Daylight Time",
	  "Africa/Khartoum" },
	{ "Libya Standard Time", "Libya Standard Time", "(UTC+02:00) Tripoli", "Libya Daylight Time",
	  "Africa/Tripoli" },
	{ "Namibia Standard Time", "Namibia Standard Time", "(UTC+02:00) Windhoek",
	  "Namibia Daylight Time", "Africa/Windhoek" },
	{ "Jordan Standard Time", "Jordan Standard Time", "(UTC+03:00) Amman", "Jordan Daylight Time",
	  "Asia/Amman" },
	{ "Arabic Standard Time", "Arabic Standard Time", "(UTC+03:00) Baghdad", "Arabic Daylight Time",
	  "Asia/Baghdad" },
	{ "Syria Standard Time", "Syria Standard Time", "(UTC+03:00) Damascus", "Syria Daylight Time",
	  "Asia/Damascus" },
	{ "Turkey Standard Time", "Turkey Standard Time", "(UTC+03:00) Istanbul",
	  "Turkey Daylight Time", "Europe/Istanbul" },
	{ "Arab Standard Time", "Arab Standard Time", "(UTC+03:00) Kuwait, Riyadh",
	  "Arab Daylight Time", "Asia/Riyadh" },
	{ "Belarus Standard Time", "Belarus Standard Time", "(UTC+03:00) Minsk",
	  "Belarus Daylight Time", "Europe/Minsk" },
	{ "Russian Standard Time", "Russia TZ 2 Standard Time", "(UTC+03:00) Moscow, St. Petersburg",
	  "Russia TZ 2 Daylight Time", "Europe/Moscow" },
	{ "E. Africa Standard Time", "E. Africa Standard Time", "(UTC+03:00) Nairobi",
	  "E. Africa Daylight Time", "Africa/Nairobi" },
	{ "Volgograd Standard Time", "Volgograd Standard Time", "(UTC+03:00) Volgograd",
	  "Volgograd Daylight Time", "Europe/Volgograd" },
	{ "Iran Standard Time", "Iran Standard Time", "(UTC+03:30) Tehran", "Iran Daylight Time",
	  "Asia/Tehran" },
	{ "Arabian Standard Time", "Arabian Standard Time", "(UTC+04:00) Abu Dhabi, Muscat",
	  "Arabian Daylight Time", "Asia/Dubai" },
	{ "Astrakhan Standard Time", "Astrakhan Standard Time", "(UTC+04:00) Astrakhan, Ulyanovsk",
	  "Astrakhan Daylight Time", "Europe/Astrakhan" },
	{ "Azerbaijan Standard Time", "Azerbaijan Standard Time", "(UTC+04:00) Baku",
	  "Azerbaijan Daylight Time", "Asia/Baku" },
	{ "Russia Time Zone 3", "Russia TZ 3 Standard Time", "(UTC+04:00) Izhevsk, Samara",
	  "Russia TZ 3 Daylight Time", "Europe/Samara" },
	{ "Mauritius Standard Time", "Mauritius Standard Time", "(UTC+04:00) Port Louis",
	  "Mauritius Daylight Time", "Indian/Mauritius" },
	{ "Saratov Standard Time", "Saratov Standard Time", "(UTC+04:00) Saratov",
	  "Saratov Daylight Time", "Europe/Saratov" },
	{ "Georgian Standard Time", "Georgian Standard Time", "(UTC+04:00) Tbilisi",
	  "Georgian Daylight Time", "Asia/Tbilisi" },
	{ "Caucasus Standard Time", "Caucasus Standard Time", "(UTC+04:00) Yerevan",
	  "Caucasus Daylight Time", "Asia/Yerevan" },
	{ "Afghanistan Standard Time", "Afghanistan Standard Time", "(UTC+04:30) Kabul",
	  "Afghanistan Daylight Time", "Asia/Kabul" },
	{ "West Asia Standard Time", "West Asia Standard Time", "(UTC+05:00) Ashgabat, Tashkent",
	  "West Asia Daylight Time", "Asia/Tashkent" },
	{ "Qyzylorda Standard Time", "Qyzylorda Standard Time", "(UTC+05:00) Astana",
	  "Qyzylorda Daylight Time", "Asia/Qyzylorda" },
	{ "Ekaterinburg Standard Time", "Russia TZ 4 Standard Time", "(UTC+05:00) Ekaterinburg",
	  "Russia TZ 4 Daylight Time", "Asia/Yekaterinburg" },
	{ "Pakistan Standard Time", "Pakistan Standard Time", "(UTC+05:00) Islamabad, Karachi",
	  "Pakistan Daylight Time", "Asia/Karachi" },
	{ "India Standard Time", "India Standard Time",
	  "(UTC+05:30) Chennai, Kolkata, Mumbai, New Delhi", "India Daylight Time", "Asia/Calcutta" },
	{ "Sri Lanka Standard Time", "Sri Lanka Standard Time", "(UTC+05:30) Sri Jayawardenepura",
	  "Sri Lanka Daylight Time", "Asia/Colombo" },
	{ "Nepal Standard Time", "Nepal Standard Time", "(UTC+05:45) Kathmandu", "Nepal Daylight Time",
	  "Asia/Katmandu" },
	{ "Central Asia Standard Time", "Central Asia Standard Time", "(UTC+06:00) Bishkek",
	  "Central Asia Daylight Time", "Asia/Almaty" },
	{ "Bangladesh Standard Time", "Bangladesh Standard Time", "(UTC+06:00) Dhaka",
	  "Bangladesh Daylight Time", "Asia/Dhaka" },
	{ "Omsk Standard Time", "Omsk Standard Time", "(UTC+06:00) Omsk", "Omsk Daylight Time",
	  "Asia/Omsk" },
	{ "Myanmar Standard Time", "Myanmar Standard Time", "(UTC+06:30) Yangon (Rangoon)",
	  "Myanmar Daylight Time", "Asia/Rangoon" },
	{ "SE Asia Standard Time", "SE Asia Standard Time", "(UTC+07:00) Bangkok, Hanoi, Jakarta",
	  "SE Asia Daylight Time", "Asia/Bangkok" },
	{ "Altai Standard Time", "Altai Standard Time", "(UTC+07:00) Barnaul, Gorno-Altaysk",
	  "Altai Daylight Time", "Asia/Barnaul" },
	{ "W. Mongolia Standard Time", "W. Mongolia Standard Time", "(UTC+07:00) Hovd",
	  "W. Mongolia Daylight Time", "Asia/Hovd" },
	{ "North Asia Standard Time", "Russia TZ 6 Standard Time", "(UTC+07:00) Krasnoyarsk",
	  "Russia TZ 6 Daylight Time", "Asia/Krasnoyarsk" },
	{ "N. Central Asia Standard Time", "Novosibirsk Standard Time", "(UTC+07:00) Novosibirsk",
	  "Novosibirsk Daylight Time", "Asia/Novosibirsk" },
	{ "Tomsk Standard Time", "Tomsk Standard Time", "(UTC+07:00) Tomsk", "Tomsk Daylight Time",
	  "Asia/Tomsk" },
	{ "China Standard Time", "China Standard Time",
	  "(UTC+08:00) Beijing, Chongqing, Hong Kong, Urumqi", "China Daylight Time", "Asia/Shanghai" },
	{ "North Asia East Standard Time", "Russia TZ 7 Standard Time", "(UTC+08:00) Irkutsk",
	  "Russia TZ 7 Daylight Time", "Asia/Irkutsk" },
	{ "Singapore Standard Time", "Malay Peninsula Standard Time",
	  "(UTC+08:00) Kuala Lumpur, Singapore", "Malay Peninsula Daylight Time", "Asia/Singapore" },
	{ "W. Australia Standard Time", "W. Australia Standard Time", "(UTC+08:00) Perth",
	  "W. Australia Daylight Time", "Australia/Perth" },
	{ "Taipei Standard Time", "Taipei Standard Time", "(UTC+08:00) Taipei", "Taipei Daylight Time",
	  "Asia/Taipei" },
	{ "Ulaanbaatar Standard Time", "Ulaanbaatar Standard Time", "(UTC+08:00) Ulaanbaatar",
	  "Ulaanbaatar Daylight Time", "Asia/Ulaanbaatar" },
	{ "Aus Central W. Standard Time", "Aus Central W. Standard Time", "(UTC+08:45) Eucla",
	  "Aus Central W. Daylight Time", "Australia/Eucla" },
	{ "Transbaikal Standard Time", "Transbaikal Standard Time", "(UTC+09:00) Chita",
	  "Transbaikal Daylight Time", "Asia/Chita" },
	{ "Tokyo Standard Time", "Tokyo Standard Time", "(UTC+09:00) Osaka, Sapporo, Tokyo",
	  "Tokyo Daylight Time", "Asia/Tokyo" },
	{ "North Korea Standard Time", "North Korea Standard Time", "(UTC+09:00) Pyongyang",
	  "North Korea Daylight Time", "Asia/Pyongyang" },
	{ "Korea Standard Time", "Korea Standard Time", "(UTC+09:00) Seoul", "Korea Daylight Time",
	  "Asia/Seoul" },
	{ "Yakutsk Standard Time", "Russia TZ 8 Standard Time", "(UTC+09:00) Yakutsk",
	  "Russia TZ 8 Daylight Time", "Asia/Yakutsk" },
	{ "Cen. Australia Standard Time", "Cen. Australia Standard Time", "(UTC+09:30) Adelaide",
	  "Cen. Australia Daylight Time", "Australia/Adelaide" },
	{ "AUS Central Standard Time", "AUS Central Standard Time", "(UTC+09:30) Darwin",
	  "AUS Central Daylight Time", "Australia/Darwin" },
	{ "E. Australia Standard Time", "E. Australia Standard Time", "(UTC+10:00) Brisbane",
	  "E. Australia Daylight Time", "Australia/Brisbane" },
	{ "AUS Eastern Standard Time", "AUS Eastern Standard Time",
	  "(UTC+10:00) Canberra, Melbourne, Sydney", "AUS Eastern Daylight Time", "Australia/Sydney" },
	{ "West Pacific Standard Time", "West Pacific Standard Time", "(UTC+10:00) Guam, Port Moresby",
	  "West Pacific Daylight Time", "Pacific/Port_Moresby" },
	{ "Tasmania Standard Time", "Tasmania Standard Time", "(UTC+10:00) Hobart",
	  "Tasmania Daylight Time", "Australia/Hobart" },
	{ "Vladivostok Standard Time", "Russia TZ 9 Standard Time", "(UTC+10:00) Vladivostok",
	  "Russia TZ 9 Daylight Time", "Asia/Vladivostok" },
	{ "Lord Howe Standard Time", "Lord Howe Standard Time", "(UTC+10:30) Lord Howe Island",
	  "Lord Howe Daylight Time", "Australia/Lord_Howe" },
	{ "Bougainville Standard Time", "Bougainville Standard Time", "(UTC+11:00) Bougainville Island",
	  "Bougainville Daylight Time", "Pacific/Bougainville" },
	{ "Russia Time Zone 10", "Russia TZ 10 Standard Time", "(UTC+11:00) Chokurdakh",
	  "Russia TZ 10 Daylight Time", "Asia/Srednekolymsk" },
	{ "Magadan Standard Time", "Magadan Standard Time", "(UTC+11:00) Magadan",
	  "Magadan Daylight Time", "Asia/Magadan" },
	{ "Norfolk Standard Time", "Norfolk Standard Time", "(UTC+11:00) Norfolk Island",
	  "Norfolk Daylight Time", "Pacific/Norfolk" },
	{ "Sakhalin Standard Time", "Sakhalin Standard Time", "(UTC+11:00) Sakhalin",
	  "Sakhalin Daylight Time", "Asia/Sakhalin" },
	{ "Central Pacific Standard Time", "Central Pacific Standard Time",
	  "(UTC+11:00) Solomon Is., New Caledonia", "Central Pacific Daylight Time",
	  "Pacific/Guadalcanal" },
	{ "Russia Time Zone 11", "Russia TZ 11 Standard Time",
	  "(UTC+12:00) Anadyr, Petropavlovsk-Kamchatsky", "Russia TZ 11 Daylight Time",
	  "Asia/Kamchatka" },
	{ "New Zealand Standard Time", "New Zealand Standard Time", "(UTC+12:00) Auckland, Wellington",
	  "New Zealand Daylight Time", "Pacific/Auckland" },
	{ "UTC+12", "UTC+12", "(UTC+12:00) Coordinated Universal Time+12", "UTC+12", "Etc/GMT-12" },
	{ "Fiji Standard Time", "Fiji Standard Time", "(UTC+12:00) Fiji", "Fiji Daylight Time",
	  "Pacific/Fiji" },
	{ "Kamchatka Standard Time", "Kamchatka Standard Time",
	  "(UTC+12:00) Petropavlovsk-Kamchatsky - Old", "Kamchatka Daylight Time", "" },
	{ "Chatham Islands Standard Time", "Chatham Islands Standard Time",
	  "(UTC+12:45) Chatham Islands", "Chatham Islands Daylight Time", "Pacific/Chatham" },
	{ "UTC+13", "UTC+13", "(UTC+13:00) Coordinated Universal Time+13", "UTC+13", "Etc/GMT-13" },
	{ "Tonga Standard Time", "Tonga Standard Time", "(UTC+13:00) Nuku'alofa", "Tonga Daylight Time",
	  "Pacific/Tongatapu" },
	{ "Samoa Standard Time", "Samoa Standard Time", "(UTC+13:00) Samoa", "Samoa Daylight Time",
	  "Pacific/Apia" },
	{ "Line Islands Standard Time", "Line Islands Standard Time", "(UTC+14:00) Kiritimati Island",
	  "Line Islands Daylight Time", "Pacific/Kiritimati" }
};

static const size_t TimeZoneNameMapSize = ARRAYSIZE(TimeZoneNameMap);
