/* Generated from resources/FreeRDP.ico with xxd -i
 *
 * The icon must have the following properties:
 * - resolution of 256x256
 * - no alpha
 * - no alternate resolutions
 */

#ifndef FREERDP_ICO_INTERNAL_
#define FREERDP_ICO_INTERNAL_

extern const unsigned char resources_FreeRDP_ico[];
extern const unsigned int resources_FreeRDP_ico_len;

#endif /* FREERDP_ICO_INTERNAL_ */
