/* FreeRDP cursor dump to use for unit tests
 * this file was auto generated by dump_write_header
 * do not modify manually
 */

#pragma once

#include <freerdp/codec/color.h>
#include <freerdp/graphics.h>

extern const gdiPalette cursor_dump_00000003_palette;
extern const rdpPointer cursor_dump_00000003_pointer;
extern const uint8_t cursor_dump_00000003_image_bgra32[];
