/**
 * WinPR: Windows Portable Runtime
 * Debugging Utils
 *
 * Copyright 2015 Armin Novak <armin.novak@thincast.com>
 * Copyright 2015 Thincast Technologies GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <winpr/config.h>

#include <winpr/buildflags.h>

#include <stdlib.h>
#include <stdio.h>
#include <winpr/crt.h>
#include <winpr/string.h>
#include <winpr/winpr.h>
#include <winpr/version.h>
#include <winpr/wlog.h>

#if !defined(WIN32)
#include <pthread.h>
#endif

void winpr_get_version(int* major, int* minor, int* revision)
{
	if (major)
		*major = WINPR_VERSION_MAJOR;
	if (minor)
		*minor = WINPR_VERSION_MINOR;
	if (revision)
		*revision = WINPR_VERSION_REVISION;
}

const char* winpr_get_version_string(void)
{
	return WINPR_VERSION_FULL;
}

const char* winpr_get_build_revision(void)
{
	return WINPR_GIT_REVISION;
}

const char* winpr_get_build_config(void)
{
	static const char build_config[] =
	    "Build configuration: " WINPR_BUILD_CONFIG "\n"
	    "Build type:          " WINPR_BUILD_TYPE "\n"
	    "CFLAGS:              " WINPR_CFLAGS "\n"
	    "Compiler:            " WINPR_COMPILER_ID ", " WINPR_COMPILER_VERSION "\n"
	    "Target architecture: " WINPR_TARGET_ARCH "\n";

	return build_config;
}
